import os, time, logging, multiprocessing
from logging.handlers import QueueHandler

class ColoredFormatter(logging.Formatter):
    COLORS = {
        'ProcessNo.0': '\033[94m',  # Light Green
        'ProcessNo.1': '\033[95m',   # Light Cyan
        'ProcessNo.2': '\033[96m',   # Light Blue
        'WARNING': '\033[93m', # Yellow
        'ERROR': '\033[91m',  # Red
        'RESET': '\033[0m'    # Reset
    }
    def format(self, record):
        color = self.COLORS.get(record.processName, self.COLORS['RESET'])
        if record.levelname == "WARNING" or record.levelname == "ERROR":
            color = self.COLORS.get(record.levelname, self.COLORS['RESET'])
        message = super().format(record)
        return f'{color}{message}{self.COLORS["RESET"]}'


def initCentralLogger(queue, rundir, loglevel):
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    console_formatter = ColoredFormatter('%(processName)-11s: %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(console_formatter)
    loglevels = {
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warning': logging.WARNING,
            'error': logging.ERROR,
            }
    console_handler.setLevel(loglevels[loglevel])
    root.addHandler(console_handler)

    file_formatter = logging.Formatter('%(asctime)s %(processName)-11s %(module)-18s %(levelname)-7s: %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')
    folder_path = os.path.join(rundir, "logs")
    os.makedirs(folder_path, exist_ok = True)
    file_handler = logging.FileHandler(os.path.join(folder_path, "eagleowl.log"))
    file_handler.setFormatter(file_formatter)
    file_handler.setLevel(logging.DEBUG)
    root.addHandler(file_handler)

    root.debug(f"{multiprocessing.current_process().name} starts.")
    while True:
        while not queue.empty():
            record = queue.get()
            if record is None:
                root.debug(f"{multiprocessing.current_process().name} ends.")
                return
            logging.getLogger(record.name).handle(record)
        time.sleep(1)

def initWorkerLogger(queue):
    root = logging.getLogger()
    if not any(isinstance(handler, QueueHandler) for handler in root.handlers):
        root.addHandler(QueueHandler(queue))
    root.setLevel(logging.DEBUG)

    pyhelm3 = logging.getLogger('pyhelm3.command')
    pyhelm3.setLevel(logging.INFO)



import os, time, json, yaml, base64, logging, traceback, multiprocessing
from abc import ABC, abstractmethod
from eagleowl.loggers import initWorkerLogger
from eagleowl.client import command, kubectl, helm

logger = logging.getLogger(__name__)

class Task(ABC):
    def __init__(self, r, resources):
        self.r = r
        self.resources = resources
        self.resource = resources[r]

    def perform(self, action):
        if action == "install_or_upgrade":
            self.install_or_upgrade()
        elif action == "uninstall":
            self.uninstall()
        self.wait_for_completion(action)
        if action == "install_or_upgrade":
            self.collect_information()

    @abstractmethod
    def install_or_upgrade(self):
        pass

    def uninstall(self):
        logger.info(f"Ingore Resource {self.r} of Type {self.resource['type']}")

    def wait_for_completion(self, action):
        for bc in self.resource.get("beacon", []):
            if "jsonpath" in bc:
                if action == "install_or_upgrade":
                    kubectl.wait_for_desire(bc)
            else:
                if action == "install_or_upgrade":
                    kubectl.wait_for_readiness(bc)
                elif action == "uninstall":
                    kubectl.wait_for_deletion(bc)

    def collect_information(self):
        for k, v in self.resource.get('supply', {}).items():
            if not isinstance(v, str):
                self.resource['supply'][k] = command.run(v['command'], self.resource.get('host'), exceptionOnFail = True)[1]

class YAMLTask(Task):
    def install_or_upgrade(self):
        if not kubectl.should_install_or_upgrade(self.resource['yamlFiles']):
            logger.info(f"Resource {self.r} is up to date.")
            return
        kubectl.install_or_upgrade(self.resource['yamlFiles'])
        logger.info(f"Resource {self.r} is installed/upgraded successfully.")

    def uninstall(self):
        kubectl.uninstall(self.resource['yamlFiles'])
        logger.info(f"Resource {self.r} not found or uninstalled successfully.")

class ChartTask(Task):
    def get_values(self, valuesFile, additionalValues):
        values = {}
        if valuesFile:
            with open(valuesFile) as fo:
                values = yaml.safe_load(fo)
        if additionalValues:
            def fill_additinal_values(additionalValues):
                res = []
                for adv in additionalValues:
                    if isinstance(adv["value"], dict):
                        _r = adv["value"]["resource"]
                        _k = adv["value"]["key"]
                        value = self.resources[_r]["supply"][_k]
                        adv['value'] = value
                    res.append(adv)
                return res
            filled_additionalValues = fill_additinal_values(additionalValues)
            for adv in filled_additionalValues:
                keys = adv['name'].split('.')
                temp = values
                for key in keys[:-1]:
                    temp = temp[key]
                temp[keys[-1]] = adv["value"]
        return values

    def install_or_upgrade(self):
        chart = self.resource["chart"]
        release = self.resource["release"]
        release['values'] = self.get_values(release['valuesFile'], release.get('additionalValues', {}))
        if not helm.should_install_or_upgrade(release, chart):
            logger.info(f"Release {release['name']} in namespace {release['namespace']} is up to date.")
            return
        helm.install_or_upgrade(release, chart)

    def uninstall(self):
        release = self.resource["release"]
        revision = helm.get_current_revision(release["name"], release["namespace"])
        if not revision:
            logger.info(f"Resource {self.r} not found.")
            return
        logger.info(f"Revision of Release {release['name']} in namespace {release['namespace']}: {revision['revision']}.")
        logger.info(f"Start to uninstall Release {release['name']} (Revision {revision['revision']}) in namespace {release['namespace']}.")
        helm.uninstall(release)
        logger.info(f"Resource {self.r} uninstalled successfully.")

class CommandTask(Task):
    def install_or_upgrade(self):
        if self.resource.get('installed') and command.runs(self.resource['installed'], self.resource.get('host')):
            logger.info(f"Resource {self.r} of Type {self.resource['type']} is done before.")
        elif self.resource.get('install'):
            command.runs(self.resource['install'], self.resource.get('host'), exceptionOnFail = True)

    def uninstall(self):
        if self.resource.get('uninstalled') and command.runs(self.resource['uninstalled'], self.resource.get('host')):
            logger.info(f"Resource {self.r} of Type {self.resource['type']} is done before.")
        elif self.resource.get('uninstall'):
            command.runs(self.resource['uninstall'], self.resource.get('host'), exceptionOnFail = True)
        else:
            super().uninstall()

class NamespaceTask(Task):
    def __init__(self, r, resources):
        super().__init__(r, resources)
        self.folder = resources[r]['yamlFolder']
        os.makedirs(self.folder, exist_ok = True)

    def install_or_upgrade(self):
        def generate_labels_string(labels):
            if not labels:
                return ""
            labels_str = "\n".join([f"        {key}: {value}" for key, value in labels.items()])
            return f"    labels:\n{labels_str}"
        labels_yaml = generate_labels_string(self.resource.get('labels'))
        namespace_yaml = f"""
apiVersion: v1
kind: Namespace
metadata:
    name: {self.resource['name']}
{labels_yaml}
""".strip()
        yaml_file = os.path.join(self.folder, f"namespace-{self.resource['name']}.yaml")
        with open(yaml_file, 'w') as fo:
            fo.write(namespace_yaml)
        if not kubectl.should_install_or_upgrade([yaml_file]):
            logger.info(f"Resource {self.r} is up to date.")
            return
        kubectl.install_or_upgrade([yaml_file])
        logger.info(f"Resource {self.r} is installed/upgraded successfully.")

class SecretTask(Task):
    def __init__(self, r, resources):
        super().__init__(r, resources)
        self.folder = resources[r]['yamlFolder']
        os.makedirs(self.folder, exist_ok = True)

    def create_docker_registry_yaml(self, secret):
        auth = base64.b64encode(f"{secret['username']}:{secret['password']}".encode("utf-8")).decode("utf-8")
        config_dict = {
            "auths": {
                secret['registry']: {
                    "username": secret['username'],
                    "password": secret['password'],
                    "email": "",
                    "auth": auth,
                }
            }
        }
        encoded_json = base64.b64encode(json.dumps(config_dict).encode("utf-8")).decode("utf-8")
        secret_yaml = f"""
apiVersion: v1
kind: Secret
metadata:
    name: {secret['name']}
    namespace: {secret['namespace']}
data:
    .dockerconfigjson: {encoded_json}
type: kubernetes.io/dockerconfigjson
""".strip()
        yaml_file = os.path.join(self.folder, f"{secret['namespace']}-{secret['name']}.yaml")
        with open(yaml_file, 'w') as fo:
            fo.write(secret_yaml)
        return yaml_file

    def create_generic_from_file_yaml(self, secret):
        with open(secret['file'], 'rb') as fo:
            file_content = fo.read()
        encoded_content = base64.b64encode(file_content).decode()
        file_name = os.path.basename(secret['file'])
        secret_yaml = f"""
apiVersion: v1
kind: Secret
metadata:
    name: {secret['name']}
    namespace: {secret['namespace']}
data:
    {file_name}: {encoded_content}
type: Opaque
""".strip()
        yaml_file = os.path.join(self.folder, f"{secret['namespace']}-{secret['name']}.yaml")
        with open(yaml_file, 'w') as fo:
            fo.write(secret_yaml)
        return yaml_file

    def install_or_upgrade(self):
        methods = {
                "docker-registry": self.create_docker_registry_yaml,
                "generic-file": self.create_generic_from_file_yaml,
                }
        yaml_file = methods[self.resource['secretType']](self.resource["secret"])
        if not kubectl.should_install_or_upgrade([yaml_file]):
            logger.info(f"Resource {self.r} is up to date.")
            return
        kubectl.install_or_upgrade([yaml_file])
        logger.info(f"Resource {self.r} is installed/upgraded successfully.")

TaskMapping = {
        'yaml': YAMLTask,
        'chart': ChartTask,
        'command': CommandTask,
        'secret': SecretTask,
        'namespace': NamespaceTask,
        }

class Executor:
    def __init__(self, shared_resources, logger_queue):
        self.resources = shared_resources
        self.lock = multiprocessing.Lock()
        self.event = multiprocessing.Event()
        self.logger_queue = logger_queue

    def start(self, action):
        try:
            initWorkerLogger(self.logger_queue)
            logger.info(f"{multiprocessing.current_process().name} starts.")
            sleep_counter = 0
            while True:
                if self.event.is_set():
                    logger.error(f"{multiprocessing.current_process().name} Exit: Exception detected in another process.")
                    return
                resource_to_execute = None
                with self.lock:
                    if all((v.get("isDone") == True or v.get("isExecuting") == True) for v in self.resources.values()):
                        break
                    for name, resource in self.resources.items():
                        if not resource.get("isExecuting") and not resource.get("isDone") and all(self.resources[dep].get("isDone") == True for dep in resource.get("dependsOn", [])):
                            resource_to_execute = name
                            self.resources[name]["isExecuting"] = True
                            break
                if not resource_to_execute:
                    if sleep_counter % 20 == 0:
                        logger.info(f"{multiprocessing.current_process().name} Waiting for valid task...")
                    time.sleep(1)
                    sleep_counter += 1
                    continue
                logger.info(f"{multiprocessing.current_process().name} works on <============ {resource_to_execute} ============>")
                TaskMapping[self.resources[resource_to_execute]['type']](resource_to_execute, self.resources).perform(action)
                with self.lock:
                    self.resources[resource_to_execute]["isDone"] = True
            logger.info(f"{multiprocessing.current_process().name} ends.")
        except Exception as e:
            logger.error(f"Receive Exception:\n{traceback.format_exc().strip()}")
            logger.error(f"{multiprocessing.current_process().name} Exit.")
            self.event.set()



import yaml, json, time, logging, subprocess

logger = logging.getLogger(__name__)

def msg(cmd, out, err):
    message = out + '\n' + err
    return ' '.join(cmd) + ": " + message.strip()

class CommandError(Exception):
    def __init__(self, cmd, result):
        super().__init__(msg(cmd, result.stdout.strip(), result.stderr.strip()))

class KubectlError(Exception):
    def __init__(self, cmd, out, err):
        super().__init__(msg(cmd, out, err))

class HelmError(Exception):
    def __init__(self, cmd, out, err):
        super().__init__(msg(cmd, out, err))

class Command:
    def run(self, cmd, host = None, exceptionOnFail = False):
        _cmd = cmd.split() if isinstance(cmd, str) else cmd
        if not host:
            cmd = _cmd
        elif host['type'] == 'deployment':
            name = f"deployment/{host['name']}"
            cmd = ["kubectl", "exec", name, "--namespace", host['namespace'], '--']
            cmd.extend(_cmd)
        elif host['type'] == 'pod':
            cmd = ["kubectl", "exec", host['name'], "--namespace", host['namespace'], '--']
            cmd.extend(_cmd)
        else:
            raise Exception(f"Unsupport host type: {host['type']}")
        logger.info(f"{' '.join(cmd)}: start...")
        if "|" in cmd:
            _cmd = ' '.join(cmd)
            result = subprocess.run(_cmd, stdout = subprocess.PIPE, stderr = subprocess.PIPE, universal_newlines = True, shell = True)
        else:
            result = subprocess.run(cmd, stdout = subprocess.PIPE, stderr = subprocess.PIPE, universal_newlines = True)
        if result.returncode != 0:
            if exceptionOnFail:
                raise CommandError(cmd, result)
            logger.info(f"{' '.join(cmd)}: return {result.returncode}.")
        else:
            logger.info(f"{' '.join(cmd)}: succeed.")
        logger.debug(f"{msg(cmd, result.stdout.strip(), result.stderr.strip())}.")
        return result.returncode, result.stdout.strip(), result.stderr.strip()

    def runs(self, cmds, host = None, exceptionOnFail = False):
        for cmd in cmds:
            if self.run(cmd, host, exceptionOnFail)[0] != 0:
                return False
        return True

class Kubectl(Command):
    def serialize_labels(self, labels):
        selectors = []
        for k, v in labels.items():
            if isinstance(v, list):
                selectors.append(f"{k} in ({','.join(v)})")
            elif isinstance(v, str):
                selectors.append(f"{k}={v}")
            else:
                raise Exception(f"Unkown Label Structure {labels}")
        return ','.join(selectors)

    def wait_for_beacon(self, beacon, wait_for):
        cmd = ['kubectl', 'wait', beacon.get('target', 'pod'), f"--for={wait_for}"]
        if 'namespace' in beacon:
            cmd.extend(['--namespace', beacon['namespace']])
        else:
            cmd.append('-A')
        if 'labels' in beacon:
            label_selector = self.serialize_labels(beacon['labels'])
            cmd.extend(['-l', label_selector])
        cmd.extend(['--timeout', beacon.get('timeout', '5m')])
        code, out, err = self.run(cmd)
        if code != 0:
            if wait_for == 'delete' and "the server doesn't have a resource type" in err:
                logger.info(f"The CRD of {beacon['target']} Not Found.")
                return
            raise KubectlError(cmd, out, err)

    def wait_for_readiness(self, beacon):
        self.wait_for_beacon(beacon, 'condition=Ready')

    def wait_for_deletion(self, beacon):
        self.wait_for_beacon(beacon, 'delete')

    def wait_for_desire(self, beacon):
        timeout = int(beacon.get('timeout', '300'))
        start_time = time.time()
        archived = False
        while time.time() - start_time < timeout:
            cmd = ["kubectl", "get", beacon["name"], "--namespace", beacon['namespace'], '-o', f"jsonpath='{beacon['jsonpath']}'"]
            code, out, err = self.run(cmd)
            if code == 0:
                state = out.strip("'")
                if isinstance(beacon["desire"], str) and state == beacon["desire"] or isinstance(beacon["desire"], list) and state in beacon["desire"]:
                    logger.info(f"{' '.join(cmd)}: achieve desired: {state}.")
                    archived = True
                    break
                logger.info(f"{' '.join(cmd)}: Current state: '{state}', continue waiting for desire state {beacon['desire']}...")
            else:
                if "NotFound" in err or "not found" in err:
                    logger.info(f"{' '.join(cmd)}: {beacon['name']} not found yet.")
                else:
                    raise KubectlError(cmd, out, err)
            time.sleep(20)
        if not archived:
            raise Exception(f"Timeout: {' '.join(cmd)}: not achieve desired: {beacon['desire']}.")

    def should_install_or_upgrade(self, yamlFiles):
        yamls = {file: 0 for file in yamlFiles}
        for f in yamls:
            cmd = ["kubectl", "diff", "-f", f]
            code, out, err = self.run(cmd)
            if code == 0:
                logger.info(f"{' '.join(cmd)}: no difference.")
            elif code == 1:
                logger.info(f"{' '.join(cmd)}: Not Found or Changed.")
                yamls[f] += 1
            else:
                raise KubectlError(cmd, out, err)
        return sum(yamls.values()) > 0

    def install_or_upgrade(self, yamlFiles):
        for f in yamlFiles:
            cmd = ['kubectl', 'apply', '-f', f]
            self.run(cmd, exceptionOnFail = True)

    def uninstall(self, yamlFiles):
        for f in yamlFiles:
            cmd = ['kubectl', 'delete', '-f', f]
            code, out, err = self.run(cmd)
            if code == 0:
                logger.info(f"{' '.join(cmd)}: uninstall successfully.")
                return
            if 'resource mapping not found' in err and 'ensure CRDs are installed first' in err:
                logger.info(f"{' '.join(cmd)}: CRD not found.")
                return
            if 'NotFound' in err:
                logger.info(f"{' '.join(cmd)}: resources not found.")
                return
            raise KubectlError(cmd, out, err)

class Helm(Command):
    def get_chart_metadata(self, chart):
        cmd = ['helm', "show", "chart", chart['ref']]
        if "version" in chart:
            cmd.extend(['--version', chart['version']])
        if "repo" in chart:
            cmd.extend(["--repo", chart["repo"]])
        code, out, err = self.run(cmd)
        if code != 0:
            raise HelmError(cmd, out, err)
        metadata = yaml.safe_load(out)
        logger.debug(f"Chart {chart['ref']} metadata retrieved.")
        return metadata

    def get_current_revision(self, release_name, namespace):
        cmd = ['helm', "status", release_name, "--namespace", namespace, "--output", "json"]
        code, out, err = self.run(cmd)
        if code != 0:
            if "not found" in err:
                logger.debug(f"Release {release_name} not Found.")
                return None
            raise HelmError(cmd, out, err)
        status = json.loads(out)
        revision = {
                "release": {
                    "name": release_name,
                    "namespace": namespace,
                    },
                "revision": status["version"],
                "status": status["info"]["status"],
                "updated": status["info"]["last_deployed"],
                "description": status["info"].get("description"),
                "notes": status["info"].get("notes"),
                }
        logger.debug(f"Current revision of {release_name}: {revision['revision']}")
        return revision

    def get_current_revision_chart_metadata(self, release_name, namespace):
        CHART_METADATA_TEMPLATE = """
name: {{ .Release.Chart.Metadata.Name }}
version: {{ .Release.Chart.Metadata.Version }}
"""
        cmd = ['helm', "get", 'all', release_name, "--namespace", namespace, "--template", '{{ .Release.Chart.Metadata.Name }} {{ .Release.Chart.Metadata.Version }}']
        code, out, err = self.run(cmd)
        if code != 0:
            raise HelmError(cmd, out, err)
        chart_metadata = {"name": out.split()[0], "version": out.split()[1]}
        logger.debug(f"The chart in current revision: {chart_metadata['name']} {chart_metadata['version']}")
        return chart_metadata

    def get_current_revision_values(self, release_name, namespace):
        cmd = ['helm', "get", 'values', release_name, "--namespace", namespace, "--output", "json"]
        code, out, err = self.run(cmd)
        if code != 0:
            raise HelmError(cmd, out, err)
        values = json.loads(out)
        logger.debug(f"The values in current revision: {values}")
        return values

    def should_install_or_upgrade(self, release, chart):
        logger.debug(f"Check Whether Release {release['name']} in namespace {release['namespace']} needs to upgrade.")

        revision = self.get_current_revision(release["name"], release["namespace"])
        if not revision:
            logger.info(f"Release {release['name']} not found")
            return True
        if revision["status"] != "deployed":
            logger.info(f"The Status of Release {release['name']}: {revision['status']}")
            return True

        revision['chart'] = self.get_current_revision_chart_metadata(release["name"], release["namespace"])
        chart['metadata'] = self.get_chart_metadata(chart)
        if revision['chart']['name'] != chart['metadata']['name']:
            logger.info(f"Old Chart: {revision['chart']['name']}, New Chart: {chart['metadata']['name']}")
            return True
        if revision['chart']['version'] != chart['metadata']['version']:
            logger.info(f"Old Version: {revision['chart']['name']}, New Version: {chart['metadata']['version']}")
            return True

        revision["values"] = self.get_current_revision_values(release["name"], release["namespace"])
        if revision["values"] != release["values"]:
            logger.debug(f"Old Values:\n{revision['values']}")
            logger.debug(f"New Values:\n{release['values']}")
            logger.info("Values Changed")
            return True

        logger.debug(f"Release {release['name']} is up to date.")
        return False

    def install_or_upgrade(self, release, chart):
        with open(release["valuesFile"], 'w') as fo:
            yaml.dump(release["values"], fo, default_flow_style=False, sort_keys=False, indent=2)
        cmd = ["helm", "upgrade", "--install", release['name'], chart['ref'], '--values', release['valuesFile'], "--namespace", release['namespace'], "--create-namespace", '--wait', "--atomic", '--timeout', "5m"]
        if 'version' in chart:
            cmd.extend(["--version", chart['version']])
        if 'repo' in chart:
            cmd.extend(['--repo', chart['repo']])
        self.run(cmd, exceptionOnFail = True)
        logger.info(f"Successfully install/upgrade Release {release['name']} of Chart {chart['ref']}.")

    def uninstall(self, release):
        cmd = ['helm', 'uninstall', release['name'], "--namespace", release['namespace'], '--wait', '--timeout', '5m']
        self.run(cmd, exceptionOnFail = True)
        logger.info(f"Successfully uninstall Release {release['name']} in namespace {release['namespace']}.")

command = Command()
helm = Helm()
kubectl = Kubectl()



import os, sys, shutil, argparse, logging, traceback, multiprocessing
from datetime import datetime
from eagleowl.loggers import initCentralLogger, initWorkerLogger
from eagleowl.executor import Executor
from eagleowl.manifest import Manifest
from eagleowl import __version__

class CustomAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if values:
            setattr(namespace, self.dest, values)
        else:
            setattr(namespace, self.dest, ['all'])

def main():
    parser = argparse.ArgumentParser(description = 'A tool for installing sophisticated Kubernetes application consisting of numerous components and intricate configurations.', allow_abbrev = False)
    parser.add_argument('-m', "--manifest", dest = 'workdir', default = "", type = str, help = 'the Application Manifest Directory, if not specified, use The Current Working Directory.')
    parser.add_argument('-r', "--rundir", default = "", type = str, help = 'the Directory containing runtime files, if not specified, use "Application-Manifest-Directory/runs/TIMESTAMP".')
    parser.add_argument('-l', "--loglevel", default = "info", choices = ['debug', 'info', 'warning', 'error'], help = 'verbose Level')
    parser.add_argument('-s', '--set', action='append', type=lambda kv: kv.split("="), dest='settings', help='overwrite fields in Application Manifest')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-i', "--install", "--upgrade", "--install_or_upgrade", dest = "install_or_upgrade", nargs='*', action = CustomAction, help = 'install/upgrade Resource(s)')
    group.add_argument('-u', "--uninstall", nargs = '*', action = CustomAction, help = 'uninstall Resource(s)')
    group.add_argument('-g', "--graph", help=argparse.SUPPRESS, action = "store_true")
    parser.add_argument('-v', "--version", action = "store_true", help = 'print the Version of EagelOwl')
    args = parser.parse_args()

    if len(sys.argv) == 1 or not (args.install_or_upgrade or args.uninstall or args.graph or args.version):
        parser.print_help()
        parser.exit()

    if args.version:
        print(__version__)
        return

    if not args.workdir:
        args.workdir = os.getcwd()
    if not args.rundir:
        args.rundir = os.path.join(args.workdir, 'runs', datetime.now().strftime("%Y%m%d-%H-%M-%S"))
        os.makedirs(args.rundir, exist_ok=True)

    logger_queue = multiprocessing.Queue(-1)
    logger_process = multiprocessing.Process(target = initCentralLogger, args = (logger_queue, args.rundir, args.loglevel), name = "LogProcess")
    logger_process.start()

    try:
        code = 1
        initWorkerLogger(logger_queue)
        logger = logging.getLogger(__name__)
        logger.info(f"{multiprocessing.current_process().name} starts.")
        logger.info(f"Working Directory: {args.workdir}.")

        if not shutil.which("helm"):
            logger.error("Cannot find 'helm' executable")
            return
        if not shutil.which("kubectl"):
            logger.error("Cannot find 'kubectl' executable")
            return

        manifest = Manifest(args.workdir, args.rundir, args.settings)
        if not manifest.initialise():
            return

        if args.graph:
            manifest.print_graph()
            code = 0
            return

        if args.uninstall:
            action = "uninstall"
            manifest.reverse_dependencies()
            manifest.set_targets(args.uninstall)
        else:
            action = "install_or_upgrade"
            manifest.set_targets(args.install_or_upgrade)

        executor = Executor(manifest.generate_shared_resources(), logger_queue)
        processes = [multiprocessing.Process(target = executor.start, args = (action,), name = f"ProcessNo.{i}") for i in range(3)]
        for process in processes:
            process.start()
        for process in processes:
            process.join()
        if not executor.event.is_set():
            code = 0
    except Exception as e:
        logger.error(f"Receive Exception:\n{traceback.format_exc().strip()}")
    finally:
        logger.info(f"{multiprocessing.current_process().name} ends.")
        logger.info(f"Detailed Logs: {os.path.join(args.rundir, 'log', 'eagleowl.log')}")
        logger_queue.put(None)
        logger_process.join()
        sys.exit(code)

if __name__ == '__main__':
    main()
